"""

	Based on ... below (heavily modified by David Brayshaw, UoR, Sep 2023)

 Sample python script for reading NetCDF files in the Gonzalez et al. 2020 dataset
 Sample plotting example
 
Hannah Bloomfield - UoR - Jan 2021


Example loading script for European weather and energy data

Once you run this you will have both hindcast and ERA5 reanalysis variables 
loaded in for matching dates. 
Country names and forecast date are also loaded and converted into a useable 
format.


"""

import numpy as np
from netCDF4 import Dataset
import matplotlib.pyplot as plt

def s2s4e_checkvarnames(filename):
	# load in the .nc file
	dataset = Dataset(filename)

	# this command gives the short names of the .nc file contents
	print(dataset.variables.keys())
	
	dataset.close()
	
	

def s2s4e_loadsingle(datadir, filename, var):


	# load in the .nc file
	dataset = Dataset(datadir + filename)

	# this command gives the short names of the .nc file contents
	dataset.variables.keys()

	# load the forecast. The array dimensions are: 
	# [country, day , year , fcst start , ensemble member ]
	fc = dataset.variables[var+'_fcst'][:]
	# load the observaions. The array dimensions are: 
	# [country, day , year , fcst start ] 
	obs = dataset.variables[var+'_rea'][:]

	# load the dates that each of the forecasts are valid for. The array dimensions are: 
	# [day , year , fcst start ] 
	# Units = days since 01,01,1970, the same as in ERA5.
	date = dataset.variables['date'][:]

	# load the identifying codes for each country.
	country_codes = dataset.variables['country_codes'][:]

	dataset.close()


	# convert the date handle into something more useful
	# set the initial date as the '0' date in the dates.
	initial_date = np.datetime64('1970-01-01')
	#setup a nump.datetime array to store the new dates in
	datetime_array = np.zeros(np.shape(date),dtype='datetime64[D]')

	#loop through the dimensions and assign the dates in the array
	for day_i in range(0,44):
		for year_i in range(0,20):
			for start_i in range(0,8):

				datetime_array[day_i, year_i, start_i] = initial_date + \
				np.timedelta64(np.int(date[day_i, year_i, start_i]),'D')


	# convert the country codes into country names using a dictionary
	country_dictionary = {
    	country_codes[0][0] + country_codes[0][1] : 'Austria',
    	country_codes[1][0] + country_codes[1][1] : 'Belgium',
    	country_codes[2][0] + country_codes[2][1] : 'Bulgaria',
    	country_codes[3][0] + country_codes[3][1] : 'Croatia',
    	country_codes[4][0] + country_codes[4][1] : 'Czech Republic',
    	country_codes[5][0] + country_codes[5][1] : 'Denmark',
    	country_codes[6][0] + country_codes[6][1] : 'Finland',
    	country_codes[7][0] + country_codes[7][1] : 'France',
    	country_codes[8][0] + country_codes[8][1] : 'Germany',
    	country_codes[9][0] + country_codes[9][1] : 'Greece',
    	country_codes[10][0] + country_codes[10][1] : 'Hungary',
    	country_codes[11][0] + country_codes[11][1] : 'Ireland',
    	country_codes[12][0] + country_codes[12][1] : 'Italy',
    	country_codes[13][0] + country_codes[13][1] : 'Latvia',
    	country_codes[14][0] + country_codes[14][1] : 'Lithuania',
    	country_codes[15][0] + country_codes[15][1] : 'Luxembourg',
    	country_codes[16][0] + country_codes[16][1] : 'Montenegro',
    	country_codes[17][0] + country_codes[17][1] : 'Netherlands',
    	country_codes[18][0] + country_codes[18][1] : 'Norway',
    	country_codes[19][0] + country_codes[19][1] : 'Poland',
    	country_codes[20][0] + country_codes[20][1] : 'Portugal',
    	country_codes[21][0] + country_codes[21][1] : 'Romania',
    	country_codes[22][0] + country_codes[22][1] : 'Slovakia',
    	country_codes[23][0] + country_codes[23][1] : 'Slovenia',
    	country_codes[24][0] + country_codes[24][1] : 'Spain',
    	country_codes[25][0] + country_codes[25][1] : 'Sweden',
    	country_codes[26][0] + country_codes[26][1] : 'Switzerland',
    	country_codes[27][0] + country_codes[27][1] : 'United Kingdom',

	}

	# create an empty list to assign country names to.
	country_full_names = []
	# assign country names
	for count_i in range(0,len(country_codes)):
		country_full_names.append( country_dictionary[country_codes[count_i][0] + 
                                                  	country_codes[count_i][1]])


	return fc, obs, datetime_array, country_full_names




# Extract a lead window from a set of s2s4e forecasts
def s2s4e_extractleadwindow(fc, obs, datetime_array, vstart, vend):

	# Notes om indicies
	# datetime_array[i,j,k] : i = lead time (day 0, 1, 2, 3, ...)
	#                         j = year (1996, 1997, 1998, ...)
	#                         k = kth launch in year (1st, 2nd, 3rd, ...)

	# Obs is in the form: [country, leadtime, year, fcst start date]
	# so we want to average over a given set of lead-times
	obs_lead = np.mean(obs[:,vstart:vend+1,:,:],1)

	# Ditto for fc, but here we have multiple fc members
	fc_lead = np.mean(fc[:,vstart:vend+1,:,:,:],1)

	# Lets keep our validity date for reference later (first day of window)
	valid_datetime_lead = datetime_array[vstart,:,:]

	# And our launch dates
	launch_datetime_lead = datetime_array[0,:,:]

	return fc_lead, obs_lead, valid_datetime_lead, launch_datetime_lead



# Load in a set of s2s4e forecasts and extract a lead window
def s2s4e_getfc(datadir,fileroot,season,var,vstart,vend):

	# load in all the months in the season required
	for iseas in season:
		filename = fileroot+iseas+'.nc'
		(fc_i,obs_i,dtarr_i,country_i) = s2s4e_loadsingle(datadir,filename,var)
		if (iseas == season[0]):
			fc = fc_i
			obs = obs_i
			dtarr = dtarr_i
			country = country_i
		else:
			fc = np.append(fc,fc_i,3)
			obs = np.append(obs, obs_i,3)
			dtarr = np.append(dtarr, dtarr_i,2)
		
	# extract a particlar validity time window of interest
	(window_fc, window_obs, validstart_dtarr, launch_dtarr) = s2s4e_extractleadwindow(fc, obs, dtarr, vstart, vend)

	return window_fc, window_obs, validstart_dtarr, launch_dtarr, country






def s2s4e_example():
	
	# Use season as NDJFM
	season = ['11','12','01','02','03']

	# the directory where the data is stored in - change to match your location!
	datadir = '/scratch2/sws06djb/s2s4e/country_daily_2m_temperature/'
	fileroot = 'ecmwf_s2s_coun_t2m_fcst_w_VI_and_rea_DAILY_stmon_'
	s2s4e_checkvarnames(datadir+fileroot+'01.nc')
	var = 't2m'

	# Extract a week-2 forecast (average days 7-13 of forecast)
	vstart = 7
	vend = 13

	fc,obs,valid,launch,country = s2s4e_getfc(datadir,fileroot,season,var,vstart,vend)


	import pdb; pdb.set_trace()


