import matplotlib.pyplot as plt
import numpy as np
import scipy
import scipy.optimize as optimize
import scipy.stats as stats
from random import *
from collections import Counter
import datetime
import cvxopt as cv
from cvxopt.modeling import op, variable, dot, matrix
from netCDF4 import Dataset as ncfile
import pdb
from batch_modelcore import *
from batch_io import *
from batch_s2s4e_data import *
import properscoring as ps


# Path to where the plots will appear
path='../figs/'

# Details about weather forecast input to use
# Here just using Nov-Mar forecasts from ECMWF, 2m temperature
season = ['11','12','01','02','03']
fileroot = 'ecmwf_s2s_coun_t2m_fcst_w_VI_and_rea_DAILY_stmon_'
datadir = '/storage/silver/metstudent/ug/bu804398/s2s4e/country_daily_2m_temperature/' # your copy of the s2s4e data you want to use
s2s4e_checkvarnames(datadir+fileroot+'01.nc') # this prints the variable names in the file - you can use it to check the next line
var = 't2m' # weather variable - just the first bit, don't include _rea and _fcst

# Parameters to investigate, model will loop over your choices with all choices
# in a single output file

# Met forecast elements
country= 'Austria' # ['Austria', 'Belgium', 'Bulgaria', 'Croatia', 'Czech Republic', 'Denmark', 'Finland', 'France', 'Germany', 'Greece', 'Hungary', 'Ireland', 'Italy', 'Latvia', 'Lithuania', 'Luxembourg', 'Montenegro', 'Netherlands', 'Norway', 'Poland', 'Portugal', 'Romania', 'Slovakia', 'Slovenia', 'Spain', 'Sweden', 'Switzerland', 'United Kingdom']
leadtime = 7 # e.g., 0,7,14,21 days: how far ahead are we forecasting
windowlength = 7 # e.g., 0,7,14,21: what is the the length of the forecast window (e.g., weekly-average)
variability = 1. # the forecast and obs are re-scaled to have a common stddev, this sets how variable it is
baserate = 5. # background value to 'add' to forecast (mean state)


# Decision model parameters
maxcap=4. # max capacity limit on reservoir before overtop
drawprice=0.5 # income gained per unit of drawdown
maxdraw=10. # maximum drawdown of reservoir in one timestep
mindraw=0. # minimum drawdown of reservoir in one timestep
mincap=0. # min capacity limit on reservoir before shortfall
shortprice=1. # cost per unit of shortfall
overtopprice=1. # cost per unit of overtopping
initres=0. # initial content of reservior

# Random number seed (selects the 'weather' in the Weigel model)
seed = 1

# Perform some initial set up actions (don't change this bit!)
cv.solvers.options['show_progress'] = False # Switch off the solver verbosity

# Load in a set of fc for a given lead time / averaging window
metfc, metobs, valid, launch, clist = s2s4e_getfc(datadir,fileroot,season,var,leadtime,leadtime+windowlength)
nt = metobs[0,:,:].size # number of timesteps
nens = metfc[0,0,0,:].size # number of ensemble members

# We want the forecast to 'resemble' the artificial data
# usually used to drive the bucket model... so
# rescale the variance of the metfc & observation as desired and
# add a fixed base inflow so that forecast/obs are always positive
icountry = clist.index(country)
sc_metfc = ((metfc[icountry,:,:,:] - np.mean(metfc[icountry,:,:,:])) /
							np.std(metfc[icountry,:,:,:] - np.mean(metfc[icountry,:,:,:]))) * variability + baserate
sc_metobs = ((metobs[icountry,:,:] - np.mean(metobs[icountry,:,:])) /
							np.std(metobs[icountry,:,:] - np.mean(metobs[icountry,:,:]))) * variability + baserate
sc_metobs.data[sc_metobs.data < 0.] = 0.0
sc_metfc.data[sc_metfc.data < 0.] = 0.0

# Calculate the ensemble mean forecast
sc_metfc_ensmean = np.mean(sc_metfc,2)


# calculate	rmse etc for the forecasts					
met_mae = mae(np.resize(sc_metobs,(1,nt)),np.resize(sc_metfc_ensmean,(1,nt)))
met_rmse = rmse(np.resize(sc_metobs,(1,nt)),np.resize(sc_metfc_ensmean,(1,nt)))
met_crps = ps.crps_ensemble(np.resize(sc_metobs,nt),np.resize(sc_metfc,(nt,nens))).mean()
print('MAE = ',met_mae)
print('RMSE = ',met_rmse)
print('CRPS = ',met_crps)

# Run the decision-making algorithm using the forecast
# This is where the really interesting bit starts!

# First, let's run the decision model with a 'perfect' forecast
# i.e., we manage the reservoir assuming we know the next timestep's weather
# This is the 'best' management we can achieve
decfc_perfect = gen_decfc_ens(maxdraw,
															mindraw,
															maxcap,
															mincap,
															drawprice,
															shortprice,
															overtopprice,
															initres,
															np.resize(sc_metobs,nt),
															np.resize(sc_metobs,(1,nt)),
															10,
															False)

# Now, let's run the decision model with an 'imperfect' forecast
# first, ensemble mean forecast
decfc_ensmean = gen_decfc_ens(maxdraw,
															mindraw,
															maxcap,
															mincap,
															drawprice,
															shortprice,
															overtopprice,
															initres,
															np.resize(sc_metobs,nt),
															np.resize(sc_metfc_ensmean,(1,nt)),
															20,
															False)


# second, the max-mean forecast
decfc_maxmin = gen_decfc_ens(maxdraw,
													 		mindraw,
															maxcap,
															mincap,
															drawprice,
															shortprice,
															overtopprice,
															initres,
															np.resize(sc_metobs,nt),
															np.transpose(np.resize(sc_metfc,(nt,nens))),
															30,
															False)


# third, the max-mean forecast
decfc_maxmean = gen_decfc_ens(maxdraw,
													 		mindraw,
															maxcap,
															mincap,
															drawprice,
															shortprice,
															overtopprice,
															initres,
															np.resize(sc_metobs,nt),
															np.transpose(np.resize(sc_metfc,(nt,nens))),
															40,
															False)


# finally, the max-max forecast
decfc_maxmax = gen_decfc_ens(maxdraw,
													 		mindraw,
															maxcap,
															mincap,
															drawprice,
															shortprice,
															overtopprice,
															initres,
															np.resize(sc_metobs,nt),
															np.transpose(np.resize(sc_metfc,(nt,nens))),
															50,
															False)


print('rnetincome mean, stddev: ')
print('perfect = ',np.mean(decfc_perfect['rnetincome']),np.std(decfc_perfect['rnetincome']))
print('ensmean = ',np.mean(decfc_ensmean['rnetincome']),np.std(decfc_ensmean['rnetincome']))
print('maxmin = ',np.mean(decfc_maxmin['rnetincome']),np.std(decfc_maxmin['rnetincome']))
print('maxmean = ',np.mean(decfc_maxmean['rnetincome']),np.std(decfc_maxmean['rnetincome']))
print('maxmax = ',np.mean(decfc_maxmax['rnetincome']),np.std(decfc_maxmax['rnetincome']))



# Let's now do some simple plots to illustrate what we've created
# Start by setting up some basics for convenience
# timestep is arbitary... all launches/years are 'just' forecasts at this point
# transpose/reshape is to just rearrange the forecast into a sensible shape
# i.e., from [launch, year, ensmember] to [ensmember, timestep]
timestep = np.arange(0,nt,1)
sct_metfc=np.transpose(np.resize(sc_metfc,(nt,nens)))
sct_metfc_ensmean = np.resize(sc_metfc_ensmean,nt)
sct_metobs = np.resize(sc_metobs,nt)

# Time series of the meteorological forecast
fig = plt.figure(figsize=(6, 12))

ax1 = fig.add_subplot(2,1,1)
xvals = timestep
minx = 0
maxx = 50
stepx = 10
miny = 0.
maxy = 15.01
stepy = 5.
ax1.axis([minx,maxx,miny,maxy])
ax1.tick_params(direction='out', which='both', labelsize='x-large')
ax1.set_xlabel('Timestep', fontsize='x-large')
ax1.set_ylabel('Metvar', fontsize='x-large')
ax1.set_xticks(np.arange(minx,maxx,stepx))
ax1.set_yticks(np.arange(miny,maxy,stepy))
ax1.plot(xvals,sct_metfc[0,:],label='metfc ens member',color="red",linewidth=1.)
for ii in np.arange(1,nens): ax1.plot(xvals,sct_metfc[ii,:],color="red",linewidth=1.)
ax1.plot(xvals,sct_metfc_ensmean,color="purple",label='metfc ens mean', linewidth=3.)
ax1.plot(xvals,sct_metobs,color="black",label='met obs')
ax1.legend(loc='upper right', fontsize='medium')

ax2 = fig.add_subplot(2,1,2)
xvals = timestep
minx = np.min(xvals)
maxx = np.max(xvals)
stepx = 100
miny = 0.
maxy = 15.01
stepy = 5.
ax2.axis([minx,maxx,miny,maxy])
ax2.tick_params(direction='out', which='both', labelsize='x-large')
ax2.set_xlabel('Timestep', fontsize='x-large')
ax2.set_ylabel('Metvar', fontsize='x-large')
ax2.set_xticks(np.arange(minx,maxx,stepx))
ax2.set_yticks(np.arange(miny,maxy,stepy))
ax2.plot(xvals,sct_metfc_ensmean,color="purple",label='metfc ens mean', linewidth=3.)
ax2.plot(xvals,sct_metobs,color="black",label='met obs')
ax2.legend(loc='upper right', fontsize='medium')

plt.savefig(path+'singleSF_met_timeseries.png')
plt.close(fig)


# PDFs of the ensemble forecast at timesteps 0, 1, 2
#
# Note: because we know how the 'truth' is constructed (Weigel's eqn 9)
#       we can also estimate the 'true' unpredictable noise around the predictable signal
#       hence can construct 'observational' KDEs
met_kde_space = np.linspace( 0, 15, 31 )
met_singlefc0_kde = scipy.stats.kde.gaussian_kde(sct_metfc[:,0]) (met_kde_space)
met_singlefc1_kde = scipy.stats.kde.gaussian_kde(sct_metfc[:,1]) (met_kde_space)
met_singlefc2_kde = scipy.stats.kde.gaussian_kde(sct_metfc[:,2]) (met_kde_space)


fig = plt.figure(figsize=(18, 6))

ax1 = fig.add_subplot(1,3,1)
xvals = met_kde_space
minx = np.min(met_kde_space)
maxx = np.max(met_kde_space)+0.0001
stepx = 5.
miny = 0.
maxy = 0.5
stepy = .1
ax1.axis([minx,maxx,miny,maxy])
ax1.tick_params(direction='out', which='both', labelsize='x-large')
ax1.set_xlabel('Metvar', fontsize='x-large')
ax1.set_ylabel('KDE', fontsize='x-large')
ax1.set_xticks(np.arange(minx,maxx,stepx))
ax1.set_yticks(np.arange(miny,maxy,stepy))
ax1.plot(xvals,met_singlefc0_kde,color="red",label='metfc ens (time = 0)', linewidth=1.)
ax1.plot([sct_metfc_ensmean[0],sct_metfc_ensmean[0]],[miny,maxy],color="red",label='met ens mean (time = 0)', linewidth=3.)
ax1.plot([sct_metobs[0],sct_metobs[0]],[miny,maxy],color="blue",label='met obs (time = 0)', linewidth=3.)
ax1.legend(loc='upper right', fontsize='medium')

ax2 = fig.add_subplot(1,3,2)
xvals = met_kde_space
minx = np.min(met_kde_space)
maxx = np.max(met_kde_space)+0.0001
stepx = 5.
miny = 0.
maxy = 0.5
stepy = .1
ax2.axis([minx,maxx,miny,maxy])
ax2.tick_params(direction='out', which='both', labelsize='x-large')
ax2.set_xlabel('Metvar', fontsize='x-large')
ax2.set_ylabel('KDE', fontsize='x-large')
ax2.set_xticks(np.arange(minx,maxx,stepx))
ax2.set_yticks(np.arange(miny,maxy,stepy))
ax2.plot(xvals,met_singlefc1_kde,color="red",label='metfc ens (time = 1)', linewidth=1.)
ax2.plot([sct_metfc_ensmean[1],sct_metfc_ensmean[1]],[miny,maxy],color="red",label='met ens mean (time = 0)', linewidth=3.)
ax2.plot([sct_metobs[1],sct_metobs[1]],[miny,maxy],color="blue",label='met obs (time = 0)', linewidth=3.)
ax2.legend(loc='upper right', fontsize='medium')

ax3 = fig.add_subplot(1,3,3)
xvals = met_kde_space
minx = np.min(met_kde_space)
maxx = np.max(met_kde_space)+0.0001
stepx = 5.
miny = 0.
maxy = 0.5
stepy = .1
ax3.axis([minx,maxx,miny,maxy])
ax3.tick_params(direction='out', which='both', labelsize='x-large')
ax3.set_xlabel('Metvar', fontsize='x-large')
ax3.set_ylabel('KDE', fontsize='x-large')
ax3.set_xticks(np.arange(minx,maxx,stepx))
ax3.set_yticks(np.arange(miny,maxy,stepy))
ax3.plot(xvals,met_singlefc2_kde,color="red",label='metfc ens (time = 2)', linewidth=1.)
ax3.plot([sct_metfc_ensmean[2],sct_metfc_ensmean[2]],[miny,maxy],color="red",label='met ens mean (time = 0)', linewidth=3.)
ax3.plot([sct_metobs[2],sct_metobs[2]],[miny,maxy],color="blue",label='met obs (time = 0)', linewidth=3.)
ax3.legend(loc='upper right', fontsize='medium')

plt.savefig(path+'singleSF_met_pdfs.png')
plt.close(fig)




# Time series of the decision model
fig = plt.figure(figsize=(12, 12))

ax1 = fig.add_subplot(2,2,1)
xvals = timestep
minx = 0
maxx = 50
stepx = 10
miny = 0.
maxy = 15.01
stepy = 5.
ax1.axis([minx,maxx,miny,maxy])
ax1.tick_params(direction='out', which='both', labelsize='x-large')
ax1.set_xlabel('Timestep', fontsize='x-large')
ax1.set_ylabel('Decvar', fontsize='x-large')
ax1.set_xticks(np.arange(minx,maxx,stepx))
ax1.set_yticks(np.arange(miny,maxy,stepy))
ax1.plot(xvals,decfc_perfect['rnetincome'],color="black",label='realised net income (perfect fc)')
ax1.plot(xvals,decfc_perfect['rdrawdown'],color="orange",label='realised drawdown')
ax1.plot(xvals,decfc_perfect['rstore'],color="blue",label='realised storage')
ax1.plot(xvals,decfc_perfect['rovertop'],color="green",label='realised overtopping')
ax1.plot(xvals,decfc_perfect['rshortfall'],color="red",label='realised shortfall')
ax1.legend(loc='upper right', fontsize='medium')


ax2 = fig.add_subplot(2,2,2)
xvals = timestep
minx = 0
maxx = 50
stepx = 10
miny = 0.
maxy = 15.01
stepy = 5.
ax2.axis([minx,maxx,miny,maxy])
ax2.tick_params(direction='out', which='both', labelsize='x-large')
ax2.set_xlabel('Timestep', fontsize='x-large')
ax2.set_ylabel('Decvar', fontsize='x-large')
ax2.set_xticks(np.arange(minx,maxx,stepx))
ax2.set_yticks(np.arange(miny,maxy,stepy))
ax2.plot(xvals,decfc_ensmean['rnetincome'],color="black",label='realised net income (ensmean fc)')
ax2.plot(xvals,decfc_ensmean['rdrawdown'],color="orange",label='realised drawdown')
ax2.plot(xvals,decfc_ensmean['rstore'],color="blue",label='realised storage')
ax2.plot(xvals,decfc_ensmean['rovertop'],color="green",label='realised overtopping')
ax2.plot(xvals,decfc_ensmean['rshortfall'],color="red",label='realised shortfall')
ax2.legend(loc='upper right', fontsize='medium')

ax3 = fig.add_subplot(2,2,3)
xvals = timestep
minx = 0
maxx = 50
stepx = 10
miny = 0.
maxy = 15.01
stepy = 5.
ax3.axis([minx,maxx,miny,maxy])
ax3.tick_params(direction='out', which='both', labelsize='x-large')
ax3.set_xlabel('Timestep', fontsize='x-large')
ax3.set_ylabel('Decvar', fontsize='x-large')
ax3.set_xticks(np.arange(minx,maxx,stepx))
ax3.set_yticks(np.arange(miny,maxy,stepy))
ax3.plot(xvals,decfc_maxmin['rnetincome'],color="black",label='realised net income (stoch maxmin)')
ax3.plot(xvals,decfc_maxmin['rdrawdown'],color="orange",label='realised drawdown')
ax3.plot(xvals,decfc_maxmin['rstore'],color="blue",label='realised storage')
ax3.plot(xvals,decfc_maxmin['rovertop'],color="green",label='realised overtopping')
ax3.plot(xvals,decfc_maxmin['rshortfall'],color="red",label='realised shortfall')
ax3.legend(loc='upper right', fontsize='medium')

ax4 = fig.add_subplot(2,2,4)
xvals = timestep
minx = 0
maxx = 50
stepx = 10
miny = 0.
maxy = 15.01
stepy = 5.
ax4.axis([minx,maxx,miny,maxy])
ax4.tick_params(direction='out', which='both', labelsize='x-large')
ax4.set_xlabel('Timestep', fontsize='x-large')
ax4.set_ylabel('Decvar', fontsize='x-large')
ax4.set_xticks(np.arange(minx,maxx,stepx))
ax4.set_yticks(np.arange(miny,maxy,stepy))
ax4.plot(xvals,decfc_maxmean['rnetincome'],color="black",label='realised net income (stoch maxmean)')
ax4.plot(xvals,decfc_maxmean['rdrawdown'],color="orange",label='realised drawdown')
ax4.plot(xvals,decfc_maxmean['rstore'],color="blue",label='realised storage')
ax4.plot(xvals,decfc_maxmean['rovertop'],color="green",label='realised overtopping')
ax4.plot(xvals,decfc_maxmean['rshortfall'],color="red",label='realised shortfall')
ax4.legend(loc='upper right', fontsize='medium')


plt.savefig(path+'singleSF_dec_timeseries.png')
plt.close(fig)



# Scatter plot of income vs metinflow
fig = plt.figure(figsize=(12, 12))

ax1 = fig.add_subplot(2,2,1)
xvals = sct_metobs
minx = 2
maxx = 8.01
stepx = 1.
miny = 0.
maxy = 5.01
stepy = 1.
ax1.axis([minx,maxx,miny,maxy])
ax1.tick_params(direction='out', which='both', labelsize='x-large')
ax1.set_xlabel('metvalue', fontsize='x-large')
ax1.set_ylabel('netincome', fontsize='x-large')
ax1.set_xticks(np.arange(minx,maxx,stepx))
ax1.set_yticks(np.arange(miny,maxy,stepy))
ax1.plot(xvals,decfc_perfect['rnetincome'],'xk',label='(perfect fc)')
ax1.legend(loc='upper right', fontsize='medium')

ax1 = fig.add_subplot(2,2,2)
xvals = sct_metobs
minx = 2
maxx = 8.01
stepx = 1.
miny = 0.
maxy = 5.01
stepy = 1.
ax1.axis([minx,maxx,miny,maxy])
ax1.tick_params(direction='out', which='both', labelsize='x-large')
ax1.set_xlabel('metvalue', fontsize='x-large')
ax1.set_ylabel('netincome', fontsize='x-large')
ax1.set_xticks(np.arange(minx,maxx,stepx))
ax1.set_yticks(np.arange(miny,maxy,stepy))
ax1.plot(xvals,decfc_ensmean['rnetincome'],'xk',label='(ensmean fc)')
ax1.legend(loc='upper right', fontsize='medium')

ax1 = fig.add_subplot(2,2,3)
xvals = sct_metobs
minx = 2
maxx = 8.01
stepx = 1.
miny = 0.
maxy = 5.01
stepy = 1.
ax1.axis([minx,maxx,miny,maxy])
ax1.tick_params(direction='out', which='both', labelsize='x-large')
ax1.set_xlabel('metvalue', fontsize='x-large')
ax1.set_ylabel('netincome', fontsize='x-large')
ax1.set_xticks(np.arange(minx,maxx,stepx))
ax1.set_yticks(np.arange(miny,maxy,stepy))
ax1.plot(xvals,decfc_maxmin['rnetincome'],'xk',label='(maxmin)')
ax1.legend(loc='upper right', fontsize='medium')


ax1 = fig.add_subplot(2,2,4)
xvals = sct_metobs
minx = 2
maxx = 8.01
stepx = 1.
miny = 0.
maxy = 5.01
stepy = 1.
ax1.axis([minx,maxx,miny,maxy])
ax1.tick_params(direction='out', which='both', labelsize='x-large')
ax1.set_xlabel('metvalue', fontsize='x-large')
ax1.set_ylabel('netincome', fontsize='x-large')
ax1.set_xticks(np.arange(minx,maxx,stepx))
ax1.set_yticks(np.arange(miny,maxy,stepy))
ax1.plot(xvals,decfc_maxmean['rnetincome'],'xk',label='(maxmean)')
ax1.legend(loc='upper right', fontsize='medium')



plt.savefig(path+'singleSF_scatter.png')
plt.close(fig)
